from ..utils import send_request, filter_none_params
from ..constants import (
    AFFILIATE_BASIC_INFO_ENDPOINT, AFFILIATE_REFERRAL_CODE_ENDPOINT, AFFILIATE_INVITEES_ENDPOINT,
    AFFILIATE_SUB_INVITEES_ENDPOINT, AFFILIATE_SUB_AFFILIATES_ENDPOINT, AFFILIATE_INVITEES_DAILY_ENDPOINT,
    AFFILIATE_MAX_LIMIT
)

from ..exceptions import (
    BloFinAPIException, BloFinRequestException, BloFinParameterException,
    BloFinAuthException, BloFinOrderException, BloFinPositionException,
    BloFinBalanceException
)

class AffiliateAPI:
    def __init__(self, client):
        self.client = client

    def get_basic_info(self):
        return send_request('GET', AFFILIATE_BASIC_INFO_ENDPOINT, self.client.auth, authenticate=True)

    def get_referral_code(self):
        return send_request('GET', AFFILIATE_REFERRAL_CODE_ENDPOINT, self.client.auth, authenticate=True)

    def get_invitees(self, uid=None, after=None, before=None, begin=None, end=None, limit=None):
        if limit and limit > AFFILIATE_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {AFFILIATE_MAX_LIMIT}")
        params = filter_none_params(uid=uid, after=after, before=before, begin=begin, end=end, limit=limit)
        return send_request('GET', AFFILIATE_INVITEES_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_sub_invitees(self, uid=None, after=None, before=None, sub_affiliate_uid=None, sub_affiliate_level=None, begin=None, end=None, limit=None):
        if limit and limit > AFFILIATE_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {AFFILIATE_MAX_LIMIT}")
        params = filter_none_params(
            uid=uid, after=after, before=before, subAffiliateUid=sub_affiliate_uid,
            subAffiliateLevel=sub_affiliate_level, begin=begin, end=end, limit=limit
        )
        return send_request('GET', AFFILIATE_SUB_INVITEES_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_sub_affiliates(self, after=None, before=None, sub_affiliate_uid=None, sub_affiliate_level=None, begin=None, end=None, limit=None):
        if limit and limit > AFFILIATE_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {AFFILIATE_MAX_LIMIT}")
        params = filter_none_params(
            after=after, before=before, subAffiliateUid=sub_affiliate_uid,
            subAffiliateLevel=sub_affiliate_level, begin=begin, end=end, limit=limit
        )
        return send_request('GET', AFFILIATE_SUB_AFFILIATES_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_invitees_daily(self, uid=None, after=None, before=None, begin=None, end=None, limit=None):
        if limit and limit > AFFILIATE_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {AFFILIATE_MAX_LIMIT}")
        params = filter_none_params(uid=uid, after=after, before=before, begin=begin, end=end, limit=limit)
        return send_request('GET', AFFILIATE_INVITEES_DAILY_ENDPOINT, self.client.auth, params=params, authenticate=True)


import requests
from urllib.parse import urlencode
from .exceptions import BloFinRequestException, raise_api_exception, BloFinAuthException
from .constants import REST_API_URL, SERVER_TIME_ENDPOINT

def check_auth_credentials(api_key, api_secret, passphrase):
    return all([api_key, api_secret, passphrase])

def send_request(method, request_path, auth, params=None, data=None, authenticate=False):
    url = REST_API_URL + request_path
    
    if authenticate and not check_auth_credentials(auth.API_KEY, auth.API_SECRET, auth.PASSPHRASE):
        raise BloFinAuthException("API credentials are required for this method. Please provide API key, API secret, and an API passphrase.")

    try:
        if method == 'GET':
            if params:
                full_path = request_path + '?' + urlencode(params)
            else:
                full_path = request_path
            headers = auth.get_headers(full_path, method) if authenticate else {}
            response = requests.get(url, headers=headers, params=params)
        elif method == 'POST':
            headers = auth.get_headers(request_path, method, data) if authenticate else {}

            response = requests.post(url, headers=headers, json=data)
        else:
            raise BloFinRequestException(f"Unsupported HTTP method: {method}")

        response.raise_for_status()
        
        if not str(response.status_code).startswith('2'):
            raise_api_exception(response)
        
        return response.json()
    
    except requests.exceptions.RequestException as e:
        raise BloFinRequestException(f"Request failed: {str(e)}")

def get_server_time():
    response = requests.get(f"{REST_API_URL}{SERVER_TIME_ENDPOINT}")
    if response.status_code == 200:
        return int(response.json()['data']['timestamp'])
    else:
        raise Exception("Failed to get server time")

def filter_none_params(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}


from typing import Optional
from ..utils import send_request, filter_none_params
from ..constants import (
    ACCOUNT_BALANCE_ENDPOINT,
    ACCOUNT_TRANSFER_ENDPOINT,
    ACCOUNT_BILLS_ENDPOINT,
    ACCOUNT_WITHDRAWAL_HISTORY_ENDPOINT,
    ACCOUNT_DEPOSIT_HISTORY_ENDPOINT,
    ACCOUNT_TYPES,
    WITHDRAWAL_STATES,
    DEPOSIT_STATES,
    MAX_LIMIT
)

from ..exceptions import BloFinParameterException

class AccountAPI:
    def __init__(self, client):
        self.client = client

    def get_balance(self, account_type: str, currency: Optional[str] = None):
        if account_type not in ACCOUNT_TYPES:
            raise BloFinParameterException(f"Invalid account_type. Must be one of {', '.join(ACCOUNT_TYPES)}")
        
        params = filter_none_params(accountType=account_type, currency=currency)
        return send_request('GET', ACCOUNT_BALANCE_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def funds_transfer(self, currency: str, amount: float, from_account: str, to_account: str, client_id: Optional[str] = None):
        if from_account not in ACCOUNT_TYPES:
            raise BloFinParameterException(f"Invalid from_account. Must be one of {', '.join(ACCOUNT_TYPES)}")
        if to_account not in ACCOUNT_TYPES:
            raise BloFinParameterException(f"Invalid to_account. Must be one of {', '.join(ACCOUNT_TYPES)}")
        
        data = filter_none_params(
            currency=currency,
            amount=str(amount),
            fromAccount=from_account,
            toAccount=to_account,
            clientId=client_id
        )
        return send_request('POST', ACCOUNT_TRANSFER_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def get_funds_transfer_history(self, currency: Optional[str] = None, from_account: Optional[str] = None, 
                                to_account: Optional[str] = None, before: Optional[str] = None, 
                                after: Optional[str] = None, limit: Optional[int] = None):
        if from_account and from_account not in ACCOUNT_TYPES:
            raise BloFinParameterException(f"Invalid from_account. Must be one of {', '.join(ACCOUNT_TYPES)}")
        if to_account and to_account not in ACCOUNT_TYPES:
            raise BloFinParameterException(f"Invalid to_account. Must be one of {', '.join(ACCOUNT_TYPES)}")
        if limit is not None and (not isinstance(limit, int) or limit <= 0 or limit > MAX_LIMIT):
            raise BloFinParameterException(f"Limit must be a positive integer not exceeding {MAX_LIMIT}")
        
        params = filter_none_params(
            currency=currency,
            fromAccount=from_account,
            toAccount=to_account,
            before=before,
            after=after,
            limit=limit
        )
        return send_request('GET', ACCOUNT_BILLS_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_withdraw_history(self, currency: Optional[str] = None, withdraw_id: Optional[str] = None, 
                            tx_id: Optional[str] = None, state: Optional[str] = None, 
                            before: Optional[str] = None, after: Optional[str] = None, 
                            limit: Optional[int] = None):
        if state and state not in WITHDRAWAL_STATES:
            raise BloFinParameterException(f"Invalid state. Must be one of {', '.join(WITHDRAWAL_STATES)}")
        if limit is not None and (not isinstance(limit, int) or limit <= 0 or limit > MAX_LIMIT):
            raise BloFinParameterException(f"Limit must be a positive integer not exceeding {MAX_LIMIT}")
        
        params = filter_none_params(
            currency=currency,
            withdrawId=withdraw_id,
            txId=tx_id,
            state=state,
            before=before,
            after=after,
            limit=limit
        )
        return send_request('GET', ACCOUNT_WITHDRAWAL_HISTORY_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_deposit_history(self, currency: Optional[str] = None, deposit_id: Optional[str] = None, 
                            tx_id: Optional[str] = None, state: Optional[str] = None, 
                            before: Optional[str] = None, after: Optional[str] = None, 
                            limit: Optional[int] = None):
        if state and state not in DEPOSIT_STATES:
            raise BloFinParameterException(f"Invalid state. Must be one of {', '.join(DEPOSIT_STATES)}")
        if limit is not None and (not isinstance(limit, int) or limit <= 0 or limit > MAX_LIMIT):
            raise BloFinParameterException(f"Limit must be a positive integer not exceeding {MAX_LIMIT}")
        
        params = filter_none_params(
            currency=currency,
            depositId=deposit_id,
            txId=tx_id,
            state=state,
            before=before,
            after=after,
            limit=limit
        )
        return send_request('GET', ACCOUNT_DEPOSIT_HISTORY_ENDPOINT, self.client.auth, params=params, authenticate=True)


from typing import Optional
from ..utils import send_request, filter_none_params
from ..constants import (
    PUBLIC_INSTRUMENTS_ENDPOINT,
    PUBLIC_TICKERS_ENDPOINT,
    PUBLIC_ORDER_BOOK_ENDPOINT,
    PUBLIC_TRADES_ENDPOINT,
    PUBLIC_MARK_PRICE_ENDPOINT,
    PUBLIC_FUNDING_RATE_ENDPOINT,
    PUBLIC_FUNDING_RATE_HISTORY_ENDPOINT,
    PUBLIC_CANDLESTICKS_ENDPOINT,
    INST_TYPE_SWAP,
    BAR_INTERVALS,
    ORDER_BOOK_MAX_SIZE,
    TRADES_MAX_LIMIT,
    FUNDING_RATE_HISTORY_MAX_LIMIT,
    CANDLESTICK_MAX_LIMIT
)

from ..exceptions import BloFinParameterException

class PublicAPI:
    def __init__(self, client):
        self.client = client

    def get_instruments(self, inst_type: Optional[str] = None):
        if inst_type and inst_type not in [INST_TYPE_SWAP]:
            raise BloFinParameterException(f"Invalid inst_type. Must be one of {INST_TYPE_SWAP}")
        params = filter_none_params(instType=inst_type)
        return send_request('GET', PUBLIC_INSTRUMENTS_ENDPOINT, self.client.auth, params=params)

    def get_tickers(self, inst_id: Optional[str] = None):
        params = filter_none_params(instId=inst_id)
        return send_request('GET', PUBLIC_TICKERS_ENDPOINT, self.client.auth, params=params)

    def get_order_book(self, inst_id: str, size: Optional[int] = None):
        if size and size > ORDER_BOOK_MAX_SIZE:
            raise BloFinParameterException(f"Size must be less than or equal to {ORDER_BOOK_MAX_SIZE}")
        params = filter_none_params(instId=inst_id, size=size)
        return send_request('GET', PUBLIC_ORDER_BOOK_ENDPOINT, self.client.auth, params=params)

    def get_trades(self, inst_id: str, limit: Optional[int] = None):
        if limit and limit > TRADES_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {TRADES_MAX_LIMIT}")
        params = filter_none_params(instId=inst_id, limit=limit)
        return send_request('GET', PUBLIC_TRADES_ENDPOINT, self.client.auth, params=params)

    def get_mark_price(self, inst_id: str):
        params = {'instId': inst_id}
        return send_request('GET', PUBLIC_MARK_PRICE_ENDPOINT, self.client.auth, params=params)

    def get_funding_rate(self, inst_id: str):
        params = {'instId': inst_id}
        return send_request('GET', PUBLIC_FUNDING_RATE_ENDPOINT, self.client.auth, params=params)

    def get_funding_rate_history(self, inst_id: str, before: Optional[str] = None, after: Optional[str] = None, limit: Optional[int] = None):
        if limit and limit > FUNDING_RATE_HISTORY_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {FUNDING_RATE_HISTORY_MAX_LIMIT}")
        params = filter_none_params(instId=inst_id, before=before, after=after, limit=limit)
        return send_request('GET', PUBLIC_FUNDING_RATE_HISTORY_ENDPOINT, self.client.auth, params=params)

    def get_candlesticks(self, inst_id: str, bar: Optional[str] = None, after: Optional[str] = None, before: Optional[str] = None, limit: Optional[int] = None):
        if bar and bar not in BAR_INTERVALS:
            raise BloFinParameterException(f"Invalid bar. Must be one of {', '.join(BAR_INTERVALS)}")
        if limit and limit > CANDLESTICK_MAX_LIMIT:
            raise BloFinParameterException(f"Limit must be less than or equal to {CANDLESTICK_MAX_LIMIT}")
        params = filter_none_params(instId=inst_id, bar=bar, after=after, before=before, limit=limit)
        return send_request('GET', PUBLIC_CANDLESTICKS_ENDPOINT, self.client.auth, params=params)


import hmac
import base64
import time
import json
import hashlib
import uuid
import urllib
from hashlib import sha256
from .utils import get_server_time

class Auth:
    def __init__(self, api_key, api_secret, passphrase, use_server_time=False):
        self.API_KEY = api_key
        self.API_SECRET = api_secret
        self.PASSPHRASE = passphrase
        self.use_server_time = use_server_time

    def get_timestamp(self):
        if self.use_server_time:
            return get_server_time()
        return str(int(round(time.time() * 1000)))

    def generate_signature(self, timestamp, method, request_path, body='', nonce=''):
        body = "" if method == "GET" else json.dumps(body)
        
        prehash_string = f"{request_path}{method.upper()}{timestamp}{nonce}{body}"
        
        signature = hmac.new(self.API_SECRET.encode(), prehash_string.encode(), hashlib.sha256)
        hexdigest = signature.hexdigest()
        hexdigest_to_bytes = hexdigest.encode()
        return base64.b64encode(hexdigest_to_bytes).decode()

    def get_headers(self, request_path: str, method: str = 'GET', body: dict = {}):
        timestamp = self.get_timestamp()
        nonce = uuid.uuid4().hex
        
        params = {}  # Add this if you need to handle query parameters
        url_params_string = urllib.parse.urlencode(params)
        url_params_string = "?" + url_params_string if url_params_string else ""
        
        headers = {
            'ACCESS-KEY': self.API_KEY,
            'ACCESS-SIGN': self.generate_signature(timestamp, method, request_path + url_params_string, body, nonce),
            'ACCESS-TIMESTAMP': timestamp,
            'ACCESS-PASSPHRASE': self.PASSPHRASE,
            'Content-Type': 'application/json',
            'ACCESS-NONCE': nonce
        }
        return headers


from .public import PublicAPI
from .account import AccountAPI
from .trading import TradingAPI
from .affiliate import AffiliateAPI
from .user import UserAPI

__all__ = ['PublicAPI', 'AccountAPI', 'TradingAPI', 'AffiliateAPI', 'UserAPI']


from importlib.metadata import version, PackageNotFoundError
from .client import BloFinClient  # Import the BloFinClient class

try:
    __version__ = version("blofin")
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"

# Expose BloFinClient at the package level
__all__ = ['BloFinClient']


# API URLs
REST_API_URL = 'https://openapi.blofin.com'
WEBSOCKET_PUBLIC_URL = 'wss://openapi.blofin.com/ws/public'
WEBSOCKET_PRIVATE_URL = 'wss://openapi.blofin.com/ws/private'

### API Endpoints

# Account endpoints
ACCOUNT_BALANCE_ENDPOINT            = '/api/v1/asset/balances'
ACCOUNT_TRANSFER_ENDPOINT           = '/api/v1/asset/transfer'
ACCOUNT_BILLS_ENDPOINT              = '/api/v1/asset/bills'
ACCOUNT_WITHDRAWAL_HISTORY_ENDPOINT = '/api/v1/asset/withdrawal-history'
ACCOUNT_DEPOSIT_HISTORY_ENDPOINT    = '/api/v1/asset/deposit-history'

# Trade endpoints
FUTURES_ACCOUNT_BALANCE_ENDPOINT      = '/api/v1/account/balance'
ACCOUNT_POSITIONS_ENDPOINT            = '/api/v1/account/positions'
ACCOUNT_MARGIN_MODE_ENDPOINT          = '/api/v1/account/margin-mode'
ACCOUNT_SET_MARGIN_MODE_ENDPOINT      = '/api/v1/account/set-margin-mode'
ACCOUNT_POSITION_MODE_ENDPOINT        = '/api/v1/account/position-mode'
ACCOUNT_SET_POSITION_MODE_ENDPOINT    = '/api/v1/account/set-position-mode'
ACCOUNT_BATCH_LEVERAGE_INFO_ENDPOINT  = '/api/v1/account/batch-leverage-info'
ACCOUNT_SET_LEVERAGE_ENDPOINT         = '/api/v1/account/set-leverage'
TRADE_ORDER_ENDPOINT                  = '/api/v1/trade/order'
TRADE_BATCH_ORDERS_ENDPOINT           = '/api/v1/trade/batch-orders'
TRADE_ORDER_TPSL_ENDPOINT             = '/api/v1/trade/order-tpsl'
TRADE_CANCEL_ORDER_ENDPOINT           = '/api/v1/trade/cancel-order'
TRADE_CANCEL_BATCH_ORDERS_ENDPOINT    = '/api/v1/trade/cancel-batch-orders'
TRADE_CANCEL_TPSL_ENDPOINT            = '/api/v1/trade/cancel-tpsl'
TRADE_ORDERS_PENDING_ENDPOINT         = '/api/v1/trade/orders-pending'
TRADE_ORDERS_TPSL_PENDING_ENDPOINT    = '/api/v1/trade/orders-tpsl-pending'
TRADE_CLOSE_POSITION_ENDPOINT         = '/api/v1/trade/close-position'
TRADE_ORDERS_HISTORY_ENDPOINT         = '/api/v1/trade/orders-history'
TRADE_ORDERS_TPSL_HISTORY_ENDPOINT    = '/api/v1/trade/orders-tpsl-history'
TRADE_FILLS_HISTORY_ENDPOINT          = '/api/v1/trade/fills-history'

# Public API Endpoints
PUBLIC_INSTRUMENTS_ENDPOINT           = '/api/v1/market/instruments'
PUBLIC_TICKERS_ENDPOINT               = '/api/v1/market/tickers'
PUBLIC_ORDER_BOOK_ENDPOINT            = '/api/v1/market/books'
PUBLIC_TRADES_ENDPOINT                = '/api/v1/market/trades'
PUBLIC_MARK_PRICE_ENDPOINT            = '/api/v1/market/mark-price'
PUBLIC_FUNDING_RATE_ENDPOINT          = '/api/v1/market/funding-rate'
PUBLIC_FUNDING_RATE_HISTORY_ENDPOINT  = '/api/v1/market/funding-rate-history'
PUBLIC_CANDLESTICKS_ENDPOINT          = '/api/v1/market/candles'

# Affiliate API endpoints
AFFILIATE_BASIC_INFO_ENDPOINT     = '/api/v1/affiliate/basic'
AFFILIATE_REFERRAL_CODE_ENDPOINT  = '/api/v1/affiliate/referral-code'
AFFILIATE_INVITEES_ENDPOINT       = '/api/v1/affiliate/invitees'
AFFILIATE_SUB_INVITEES_ENDPOINT   = '/api/v1/affiliate/sub-invitees'
AFFILIATE_SUB_AFFILIATES_ENDPOINT = '/api/v1/affiliate/sub-affiliates'
AFFILIATE_INVITEES_DAILY_ENDPOINT = '/api/v1/affiliate/invitees/daily'

# User API endpoints
USER_API_KEY_INFO_ENDPOINT = '/api/v1/user/query-apikey'

# Miscellaneous
SERVER_TIME_ENDPOINT = '/api/v1/public/time'


# Account types
ACCOUNT_TYPE_FUNDING = "funding"
ACCOUNT_TYPE_FUTURES = "futures"
ACCOUNT_TYPE_COPY_TRADING = "copy_trading"
ACCOUNT_TYPE_EARN = "earn"
ACCOUNT_TYPE_SPOT = "spot"

ACCOUNT_TYPES = [
    ACCOUNT_TYPE_FUNDING,
    ACCOUNT_TYPE_FUTURES,
    ACCOUNT_TYPE_COPY_TRADING,
    ACCOUNT_TYPE_EARN,
    ACCOUNT_TYPE_SPOT
]

# Withdrawal states
WITHDRAWAL_STATE_WAITING = "0"
WITHDRAWAL_STATE_WITHDRAWING = "1"
WITHDRAWAL_STATE_FAILED = "2"
WITHDRAWAL_STATE_APPROVED = "3"
WITHDRAWAL_STATE_CANCELED = "4"
WITHDRAWAL_STATE_KYT = "6"
WITHDRAWAL_STATE_PROCESSING = "7"

WITHDRAWAL_STATES = [
    WITHDRAWAL_STATE_WAITING,
    WITHDRAWAL_STATE_WITHDRAWING,
    WITHDRAWAL_STATE_FAILED,
    WITHDRAWAL_STATE_APPROVED,
    WITHDRAWAL_STATE_CANCELED,
    WITHDRAWAL_STATE_KYT,
    WITHDRAWAL_STATE_PROCESSING
]

# Deposit states
DEPOSIT_STATE_PENDING = "0"
DEPOSIT_STATE_DONE = "1"
DEPOSIT_STATE_FAILED = "2"
DEPOSIT_STATE_KYT = "3"

DEPOSIT_STATES = [
    DEPOSIT_STATE_PENDING,
    DEPOSIT_STATE_DONE,
    DEPOSIT_STATE_FAILED,
    DEPOSIT_STATE_KYT
]

# Candlestick bar intervals
BAR_1M = "1m"
BAR_3M = "3m"
BAR_5M = "5m"
BAR_15M = "15m"
BAR_30M = "30m"
BAR_1H = "1H"
BAR_2H = "2H"
BAR_4H = "4H"
BAR_6H = "6H"
BAR_8H = "8H"
BAR_12H = "12H"
BAR_1D = "1D"
BAR_3D = "3D"
BAR_1W = "1W"
BAR_1MO = "1M"

BAR_INTERVALS = [
    BAR_1M, BAR_3M, BAR_5M, BAR_15M, BAR_30M,
    BAR_1H, BAR_2H, BAR_4H, BAR_6H, BAR_8H, BAR_12H,
    BAR_1D, BAR_3D, BAR_1W, BAR_1MO
]

# Instrument types
INST_TYPE_SWAP = "SWAP"

INSTRUMENT_TYPES = [INST_TYPE_SWAP]

# Order side constants
ORDER_SIDE_BUY = "buy"
ORDER_SIDE_SELL = "sell"

ORDER_SIDES = [ORDER_SIDE_BUY, ORDER_SIDE_SELL]

# Order type constants
ORDER_TYPE_MARKET = "market"
ORDER_TYPE_LIMIT = "limit"
ORDER_TYPE_POST_ONLY = "post_only"
ORDER_TYPE_FOK = "fok"
ORDER_TYPE_IOC = "ioc"

ORDER_TYPES = [ORDER_TYPE_MARKET, ORDER_TYPE_LIMIT, ORDER_TYPE_POST_ONLY, ORDER_TYPE_FOK, ORDER_TYPE_IOC]

# Margin mode constants
MARGIN_MODE_CROSS = "cross"
MARGIN_MODE_ISOLATED = "isolated"

MARGIN_MODES = [MARGIN_MODE_CROSS, MARGIN_MODE_ISOLATED]

# Position side constants
POSITION_SIDE_LONG = "long"
POSITION_SIDE_SHORT = "short"
POSITION_SIDE_NET = "net"

POSITION_SIDES = [POSITION_SIDE_LONG, POSITION_SIDE_SHORT, POSITION_SIDE_NET]

# Position mode constants
POSITION_MODE_NET = "net_mode"
POSITION_MODE_LONG_SHORT = "long_short_mode"

POSITION_MODES = [POSITION_MODE_NET, POSITION_MODE_LONG_SHORT]

# Order state constants
ORDER_STATE_LIVE = "live"
ORDER_STATE_CANCELED = "canceled"
ORDER_STATE_FILLED = "filled"
ORDER_STATE_PARTIALLY_FILLED = "partially_filled"

ORDER_STATES = [ORDER_STATE_LIVE, ORDER_STATE_CANCELED, ORDER_STATE_FILLED, ORDER_STATE_PARTIALLY_FILLED]

# Sub-affiliate levels
SUB_AFFILIATE_LEVEL_2 = "2"
SUB_AFFILIATE_LEVEL_3 = "3"
SUB_AFFILIATE_LEVEL_4 = "4"

SUB_AFFILIATE_LEVELS = [SUB_AFFILIATE_LEVEL_2, SUB_AFFILIATE_LEVEL_3, SUB_AFFILIATE_LEVEL_4]

# Default and maximum limit for pagination
DEFAULT_LIMIT = 20
MAX_LIMIT = 100

# Order book max size
ORDER_BOOK_MAX_SIZE = 100

# Trades max limit
TRADES_MAX_LIMIT = 100

# Funding rate history max limit
FUNDING_RATE_HISTORY_MAX_LIMIT = 100

# Candlestick max limit
CANDLESTICK_MAX_LIMIT = 1440

# Affiliate max limit
AFFILIATE_MAX_LIMIT = 100

# Account max limit
ACCOUNT_MAX_LIMIT = 100

# Trading max order size
TRADING_MAX_ORDER_SIZE = 1000


from typing import List, Optional, Union
from ..utils import send_request, filter_none_params
from ..constants import (
    FUTURES_ACCOUNT_BALANCE_ENDPOINT,
    ACCOUNT_POSITIONS_ENDPOINT,
    ACCOUNT_MARGIN_MODE_ENDPOINT,
    ACCOUNT_SET_MARGIN_MODE_ENDPOINT,
    ACCOUNT_POSITION_MODE_ENDPOINT,
    ACCOUNT_SET_POSITION_MODE_ENDPOINT,
    ACCOUNT_BATCH_LEVERAGE_INFO_ENDPOINT,
    ACCOUNT_SET_LEVERAGE_ENDPOINT,
    TRADE_ORDER_ENDPOINT,
    TRADE_BATCH_ORDERS_ENDPOINT,
    TRADE_ORDER_TPSL_ENDPOINT,
    TRADE_CANCEL_ORDER_ENDPOINT,
    TRADE_CANCEL_BATCH_ORDERS_ENDPOINT,
    TRADE_CANCEL_TPSL_ENDPOINT,
    TRADE_ORDERS_PENDING_ENDPOINT,
    TRADE_ORDERS_TPSL_PENDING_ENDPOINT,
    TRADE_CLOSE_POSITION_ENDPOINT,
    TRADE_ORDERS_HISTORY_ENDPOINT,
    TRADE_ORDERS_TPSL_HISTORY_ENDPOINT,
    TRADE_FILLS_HISTORY_ENDPOINT,
    ORDER_SIDES,
    ORDER_TYPES,
    MARGIN_MODES,
    POSITION_SIDES,
    POSITION_MODES,
    ORDER_STATES
)

from ..exceptions import (
    BloFinAPIException,
    BloFinRequestException,
    BloFinParameterException,
    BloFinAuthException,
    BloFinOrderException,
    BloFinPositionException,
    BloFinBalanceException
)

class TradingAPI:
    def __init__(self, client):
        self.client = client

    def get_futures_account_balance(self):
        return send_request('GET', FUTURES_ACCOUNT_BALANCE_ENDPOINT, self.client.auth, authenticate=True)

    def get_positions(self, inst_id: Optional[str] = None):
        params = {'instId': inst_id} if inst_id else {}
        return send_request('GET', ACCOUNT_POSITIONS_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_margin_mode(self):
        return send_request('GET', ACCOUNT_MARGIN_MODE_ENDPOINT, self.client.auth, authenticate=True)

    def set_margin_mode(self, margin_mode: str):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        data = {'marginMode': margin_mode}
        return send_request('POST', ACCOUNT_SET_MARGIN_MODE_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def get_position_mode(self):
        return send_request('GET', ACCOUNT_POSITION_MODE_ENDPOINT, self.client.auth, authenticate=True)

    def set_position_mode(self, position_mode: str):
        if position_mode not in POSITION_MODES:
            raise BloFinParameterException(f"Invalid position_mode. Must be one of: {', '.join(POSITION_MODES)}")
        data = {'positionMode': position_mode}
        return send_request('POST', ACCOUNT_SET_POSITION_MODE_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def get_multiple_leverage(self, inst_id: str, margin_mode: str):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        params = {'instId': inst_id, 'marginMode': margin_mode}
        return send_request('GET', ACCOUNT_BATCH_LEVERAGE_INFO_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def set_leverage(self, inst_id: str, leverage: int, margin_mode: str, position_side: Optional[str] = None):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        if position_side and position_side not in POSITION_SIDES:
            raise BloFinParameterException(f"Invalid position_side. Must be one of: {', '.join(POSITION_SIDES)}")
        data = {
            'instId': inst_id,
            'leverage': leverage,
            'marginMode': margin_mode
        }
        if position_side:
            data['positionSide'] = position_side
        return send_request('POST', ACCOUNT_SET_LEVERAGE_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def place_order(self, inst_id: str, margin_mode: str, position_side: str, side: str, order_type: str, price: float, size: float, **kwargs):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        if position_side not in POSITION_SIDES:
            raise BloFinParameterException(f"Invalid position_side. Must be one of: {', '.join(POSITION_SIDES)}")
        if side not in ORDER_SIDES:
            raise BloFinParameterException(f"Invalid side. Must be one of: {', '.join(ORDER_SIDES)}")
        if order_type not in ORDER_TYPES:
            raise BloFinParameterException(f"Invalid order_type. Must be one of: {', '.join(ORDER_TYPES)}")
        
        data = {
            'instId': inst_id,
            'marginMode': margin_mode,
            'positionSide': position_side,
            'side': side,
            'orderType': order_type,
            'price': price,
            'size': size,
            **kwargs
        }
        return send_request('POST', TRADE_ORDER_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def place_multiple_orders(self, orders: List[dict]):
        return send_request('POST', TRADE_BATCH_ORDERS_ENDPOINT, self.client.auth, data=orders, authenticate=True)

    def place_tpsl_order(self, inst_id: str, margin_mode: str, position_side: str, side: str, tp_trigger_price: float, tp_order_price: float, sl_trigger_price: float, sl_order_price: float, size: float, **kwargs):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        if position_side not in POSITION_SIDES:
            raise BloFinParameterException(f"Invalid position_side. Must be one of: {', '.join(POSITION_SIDES)}")
        if side not in ORDER_SIDES:
            raise BloFinParameterException(f"Invalid side. Must be one of: {', '.join(ORDER_SIDES)}")
        
        data = {
            'instId': inst_id,
            'marginMode': margin_mode,
            'positionSide': position_side,
            'side': side,
            'tpTriggerPrice': tp_trigger_price,
            'tpOrderPrice': tp_order_price,
            'slTriggerPrice': sl_trigger_price,
            'slOrderPrice': sl_order_price,
            'size': size,
            **kwargs
        }
        return send_request('POST', TRADE_ORDER_TPSL_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def cancel_order(self, inst_id: str, order_id: str, client_order_id: Optional[str] = None):
        data = {'instId': inst_id, 'orderId': order_id}
        if client_order_id:
            data['clientOrderId'] = client_order_id
        return send_request('POST', TRADE_CANCEL_ORDER_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def cancel_multiple_orders(self, orders: List[dict]):
        return send_request('POST', TRADE_CANCEL_BATCH_ORDERS_ENDPOINT, self.client.auth, data=orders, authenticate=True)

    def cancel_tpsl_order(self, inst_id: str, tpsl_id: str, client_order_id: Optional[str] = None):
        data = {'instId': inst_id, 'tpslId': tpsl_id, 'clientOrderId': client_order_id if client_order_id else ''}
        if client_order_id:
            data['clientOrderId'] = client_order_id
        return send_request('POST', TRADE_CANCEL_TPSL_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def get_active_orders(self, inst_id: Optional[str] = None, order_type: Optional[str] = None, state: Optional[str] = None, after: Optional[str] = None, before: Optional[str] = None, limit: Optional[int] = None):
        if order_type and order_type not in ORDER_TYPES:
            raise BloFinParameterException(f"Invalid order_type. Must be one of: {', '.join(ORDER_TYPES)}")
        if state and state not in ORDER_STATES:
            raise BloFinParameterException(f"Invalid state. Must be one of: {', '.join(ORDER_STATES)}")
        
        params = filter_none_params(instId=inst_id, orderType=order_type, state=state, after=after, before=before, limit=limit)
        return send_request('GET', TRADE_ORDERS_PENDING_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def get_active_tpsl_orders(self, inst_id: Optional[str] = None, tpsl_id: Optional[str] = None, client_order_id: Optional[str] = None, after: Optional[str] = None, before: Optional[str] = None, limit: Optional[int] = None):
        params = filter_none_params(instId=inst_id, tpslId=tpsl_id, clientOrderId=client_order_id, after=after, before=before, limit=limit)
        return send_request('GET', TRADE_ORDERS_TPSL_PENDING_ENDPOINT, self.client.auth, params=params, authenticate=True)

    def close_positions(self, inst_id: str, margin_mode: str, position_side: str, client_order_id: Optional[str] = None):
        if margin_mode not in MARGIN_MODES:
            raise BloFinParameterException(f"Invalid margin_mode. Must be one of: {', '.join(MARGIN_MODES)}")
        if position_side not in POSITION_SIDES:
            raise BloFinParameterException(f"Invalid position_side. Must be one of: {', '.join(POSITION_SIDES)}")
        
        data = {
            'instId': inst_id,
            'marginMode': margin_mode,
            'positionSide': position_side
        }
        if client_order_id:
            data['clientOrderId'] = client_order_id
        return send_request('POST', TRADE_CLOSE_POSITION_ENDPOINT, self.client.auth, data=data, authenticate=True)

    def get_order_history(self, inst_id: Optional[str] = None, order_type: Optional[str] = None, state: Optional[str] = None, after: Optional[str] = None, before: Optional[str] = None, begin: Optional[str] = None, end: Optional[str] = None, limit: Optional[int] = None):
        if order_type and order_type not in ORDER_TYPES:
            raise BloFinParameterException(f"Invalid order_type. Must be one of: {', '.join(ORDER_TYPES)}")
        if state and state not in ORDER_STATES:
            raise BloFinParameterException(f"Invalid state. Must be one of: {', '.join(ORDER_STATES)}")
        
        params = filter_none_params(instId=inst_id, orderType=order_type, state=state, after=after, before=before, begin=begin, end=end, limit=limit)
        return send_request('GET', TRADE_ORDERS_HISTORY_ENDPOINT, self.client.auth, params=params, authenticate=True)


from ..utils import send_request
from ..constants import USER_API_KEY_INFO_ENDPOINT

from ..exceptions import (
    BloFinAPIException, BloFinRequestException, BloFinParameterException,
    BloFinAuthException, BloFinOrderException, BloFinPositionException,
    BloFinBalanceException
)

class UserAPI:
    def __init__(self, client):
        self.client = client

    def get_api_key_info(self):
        return send_request('GET', USER_API_KEY_INFO_ENDPOINT, self.client.auth, authenticate=True)


from blofin.constants import REST_API_URL, WEBSOCKET_PUBLIC_URL, WEBSOCKET_PRIVATE_URL
from blofin.exceptions import (
    BloFinAPIException, BloFinRequestException, BloFinParameterException,
    BloFinAuthException, BloFinOrderException, BloFinPositionException,
    BloFinBalanceException
)
from blofin.api.public import PublicAPI
from blofin.api.account import AccountAPI
from blofin.api.trading import TradingAPI
from blofin.api.affiliate import AffiliateAPI
from blofin.api.user import UserAPI
from blofin.auth import Auth
from blofin.utils import check_auth_credentials

# Rest of the client.py content...

class BloFinClient:
    def __init__(self, api_key=None, api_secret=None, passphrase=None, use_server_time=False):
        self.auth = Auth(api_key, api_secret, passphrase, use_server_time)
        self.use_server_time = use_server_time
        self.authenticated = check_auth_credentials(api_key, api_secret, passphrase)

        if not self.authenticated:
            print("Warning: API credentials not provided. Some methods requiring authentication will not be accessible.")

        self.public    = PublicAPI(self)
        self.account   = AccountAPI(self)
        self.trading   = TradingAPI(self)
        self.affiliate = AffiliateAPI(self)
        self.user      = UserAPI(self)

    def get_server_time(self):
        return self.public.get_server_time()


class BloFinAPIException(Exception):
    def __init__(self, response):
        self.code = 0
        if isinstance(response, str):
            self.message = response
        else:
            try:
                json_res = response.json()
            except ValueError:
                self.message = f'Invalid JSON error message from BloFin: {response.text}'
            else:
                if 'code' in json_res and 'msg' in json_res:
                    self.code = json_res['code']
                    self.message = json_res['msg']
                else:
                    self.message = f'Unknown error format from BloFin: {json_res}'


    def __str__(self):
        return f'BloFinAPIException(code={self.code}): {self.message}'

class BloFinRequestException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f'BloFinRequestException: {self.message}'

class BloFinParameterException(BloFinAPIException):
    """Raised when there's an issue with the parms sent to the API"""
    pass

class BloFinAuthException(BloFinAPIException):
    """Raised when there's an auth issue"""
    pass

class BloFinOrderException(BloFinAPIException):
    """issue related to orders"""
    pass

class BloFinPositionException(BloFinAPIException):
    """issue related to positions"""
    pass

class BloFinBalanceException(BloFinAPIException):
    """raised when theres an issue related to account balance"""
    pass

def raise_api_exception(response):
    code = int(response.json().get('code', 0))
    
    if 152000 <= code < 153000:
        raise BloFinParameterException(response)
    elif 152400 <= code < 152500:
        raise BloFinAuthException(response)
    elif 102000 <= code < 103000:
        raise BloFinOrderException(response)
    elif 103000 <= code < 104000 or 110000 <= code < 111000:
        raise BloFinPositionException(response)
    elif 150000 <= code < 151000:
        raise BloFinBalanceException(response)
    else:
        raise BloFinAPIException(response)